import os
import sys
import json

from jsonschema import Draft4Validator, validate as validate_schema

HERE = os.path.dirname(os.path.abspath(__file__))

__all__ = ['validate', 'export_settings']


def open_schema(schema_name):
    return open(os.path.join(HERE, schema_name))


def read_schema(schema_name, check=True):
    raw = json.load(open_schema(schema_name))
    if check:
        Draft4Validator.check_schema(raw)
    return raw


def validate(manifest):
    """
    Validate an instance of the userapp.json schema.

    manifest can be a dict object (returned as is) or a filename.
    """
    if isinstance(manifest, basestring):
        manifest = json.load(open(manifest))
    validate_schema(manifest, read_schema('userapp.json'))
    return manifest


def export_settings(manifest):
    """
    Return a dict with settings extracted from a manifest dict.
    The string "<NO_DEFAULT>" is used for entries without default values.
    """
    config = {}
    for entry in manifest['settings']:
        if 'settings' in entry:
            # Nested settings.
            config[entry['name']] = export_settings(entry)
        else:
            config[entry['name']] = entry.get('default', '<NO_DEFAULT>')

    return config


if __name__ == "__main__":
    result = export_settings(validate(sys.argv[1]))
    print json.dumps({"settings": result}, indent=4)



import random
import logging

from toredis import Client as RedisClient

__all__ = ['ReconnectingRedis']


class ReconnectingRedis(RedisClient):

    def __init__(self, suffix=None, *args, **kwargs):
        super(ReconnectingRedis, self).__init__(*args, **kwargs)
        self._log = logging.getLogger(__name__)
        self._retry = 1
        self._name = 'redis%s' % ('-%s' % suffix if suffix else '')

    def connect(self, host, port, callback=None):
        self.host = host
        self.port = port
        self.callback = callback
        self._reconnect()

    def on_disconnect(self):
        self._log.error('%s not connected, retrying in %s', self._name, self._retry)
        self._io_loop.call_later(self._retry, self._reconnect)

    def _reconnect(self):
        try:
            super(ReconnectingRedis, self).connect(self.host, self.port, self.callback)
            self._retry = 1
        except Exception as err:
            self._retry *= 2
            self._retry += random.random()
            if self._retry > 10:
                self._retry = 1
            self._log.error('%s failed to connect to %s:%s - %s',
                    self._name, self.host, self.port, err)
        else:
            if self.is_connected():
                self._log.debug('%s connected', self._name)



from flipdisc._particle import lib, ffi

__all__ = ["Emitter"]

_EMITTER_KEYS = {
    'input_radius': float,
    'input_amplitude': float,
}
_PARTICLE_KEYS = {
    'friction': float,
    'gravity': float,
    'gravity_acceleration': float,
    'gravity_enabled': bool,
    'restitution': float,
    'offscreen_delay_max': float,
    'offscreen_delay_min': float,
    'spawn_radius_max': float,
    'spawn_radius_min': float,
    'speed': float,
}
_LIMIT_TO_01 = frozenset(['friction', 'restitution'])
_LIMIT_TO_0 = frozenset(['gravity'])


class Emitter(object):

    def __init__(self, step=1./30):
        self._ctx = lib.emitter_context()
        self._step = step
        self._elapsed = 0.0

    def __del__(self):
        lib.emitter_free(self._ctx)

    def update(self, step=None):
        if step is None:
            step = self._step
        self._elapsed += step
        self.set_elapsed_time(self._elapsed)
        lib.emitter_update(self._ctx)

    def set_elapsed_time(self, elapsed):
        pctx = self._ctx.pctx
        pctx.elapsed_time = elapsed

    def set_size(self, height, width):
        pctx = self._ctx.pctx
        pctx.size.x = height
        pctx.size.y = width

    def get_settings(self):
        pctx = self._ctx.pctx
        settings = {
            'size': {'x': pctx.size.x, 'y': pctx.size.y}
        }
        for name in _PARTICLE_KEYS:
            settings[name] = getattr(pctx, name)
        return settings

    def set_setting(self, name, value):
        if name in _EMITTER_KEYS:
            value = _EMITTER_KEYS[name](value)
            if value <= 0:
                raise ValueError('"%s" must be greater than 0')
            setattr(self._ctx, name, value)
            return

        pctx = self._ctx.pctx
        if name in _PARTICLE_KEYS:
            value = _PARTICLE_KEYS[name](value)
            if name in _LIMIT_TO_01 and (value < 0 or value > 1):
                raise ValueError('"%s" must be between 0 and 1' % name)
            if name in _LIMIT_TO_0 and value < 0:
                raise ValueError('"%s" must be greater or equal to 0' % name)
            setattr(pctx, name, value)
            return

        raise KeyError('Unknown setting "%s"' % name)

    def clear(self):
        lib.emitter_remove_all(self._ctx)

    def add_particle(self, position):
        return lib.emitter_add_particle(self._ctx, position)

    def add_force(self, position, velocity):
        lib.emitter_add_force(self._ctx, position, velocity)

    def draw(self, mat, value=255):
        height, width = mat.shape
        arr = ffi.cast('uint8_t*', ffi.from_buffer(mat))
        lib.particles_draw(self._ctx.pctx, value, arr, width, height)



import cv2
import numpy


def image_to_particles(emitter, bin_image):
    """
    Add every point from bin_image != 0 as a particle.

    * emitter should be an instance of flipdisc.particle.Emitter
    * bin_image should be a numpy 2d array
    """
    emitter.clear()
    for coord in numpy.argwhere(bin_image != 0):
        emitter.add_particle(tuple(coord))


def flip(arr, mirror, upsidedown):
    """Flips a numpy array and return it."""
    if mirror:
        arr = numpy.fliplr(arr)
    if upsidedown:
        arr = numpy.flipud(arr)
    return arr


def pad(image, size, value=0):
    """Pad and return image with border size and value specified."""
    return numpy.pad(image, size, 'constant', constant_values=value)


def resize(image, target_size):
    """
    Resize image to the target size while keeping the aspect ratio.
    A new image of size <= target_size is returned.

    * image is assumed to be a numpy 2d array with a single channel
    * target_size is assumed to be a tuple of integers (height, width)
    """
    height, width = target_size

    im_height, im_width = image.shape
    ratio_a = im_width / float(im_height)
    ratio_b = im_height / float(im_width)

    if width * ratio_b <= height:
        final_size = (width, int(width * ratio_b))
    else:
        final_size = (int(height * ratio_a), height)

    if min(target_size) > max(image.shape):
        interpolation = cv2.INTER_LINEAR
    else:
        interpolation = cv2.INTER_AREA

    return cv2.resize(image, final_size, interpolation=interpolation)


def copy_to_center(dest, img):
    """
    Copy img to the center of dest.

    * dest must be at least as big as img
    * img and dest should both be numpy arrays
    """
    dest_height, dest_width = dest.shape
    src_height, src_width = img.shape
    if src_height > dest_height or src_width > dest_width:
        raise Exception('img %s is larger than dest %s' % (img.shape, dest.shape))

    d_half = dest_height / 2, dest_width / 2
    (half_h, half_exh), (half_w, half_exw) = divmod(src_height, 2), divmod(src_width, 2)
    dest[d_half[0] - half_h:d_half[0] + half_h + half_exh,
         d_half[1] - half_w:d_half[1] + half_w + half_exw] = img



import cv2
import numpy


class OpticalFlow(object):

    def __init__(self, trackrate, max_features, min_threshold, reverse=True):
        self._trackrate = trackrate
        self._max_features = max_features
        self._min_threshold = min_threshold
        self._reverse = reverse

        self._frame_count = 0

        self._prev_frame = None
        self._prev_features = None
        self._features = None

    def update_settings(self,
            trackrate=None, max_features=None, min_threshold=None, reverse=None):
        if trackrate is not None:
            self._trackrate = trackrate
        if max_features is not None:
            self._max_features = max_features
        if min_threshold is not None:
            self._min_threshold = min_threshold
        if reverse is not None:
            self._reverse = reverse

    def update(self, frame, emitter):
        self._frame_count += 1

        feat_status = None
        if self._prev_frame is not None:
            if self._features is None or self._frame_count % self._trackrate == 0:
                self._features = cv2.goodFeaturesToTrack(self._prev_frame, self._max_features, 0.005, 3.0)
            if self._features is not None:
                self._prev_features = self._features.copy()
                self._features, feat_status, err = cv2.calcOpticalFlowPyrLK(
                        self._prev_frame, frame, self._prev_features, self._features)
        self._prev_frame = frame.copy()

        # Apply force to particles using the current optical flow info.
        height = frame.shape[0]
        features = self._features if self._features is not None else []
        feat_status = feat_status if feat_status is not None else []
        for indx, (feat, status) in enumerate(zip(features, feat_status)):
            if not status:
                continue
            destination = feat[0][::-1]
            origin = self._prev_features[indx][0][::-1]  # XXX is there a guarantee that this indx exists?
            velocity = origin - destination if self._reverse else destination - origin
            norm = numpy.sqrt((velocity ** 2).sum())
            if norm / height > self._min_threshold:
                emitter.add_force(tuple(origin), tuple(velocity))



import cv2
import numpy

from flipdisc import util

__all__ = [
    'INTER_NEAREST', 'INTER_LINEAR', 'INTER_AREA', 'INTER_CUBIC', 'INTER_LANCZOS4',
    'resize_stretch', 'resize_cover', 'resize_contain', 'load_image',
]


# A nearest-neighbor interpolation.
INTER_NEAREST = cv2.INTER_NEAREST

# A bilinear interpolation.
INTER_LINEAR  = cv2.INTER_LINEAR

# Resampling using pixel area relation.
# It may be a preferred method for image decimation,
# as it gives moire'-free results. But when the image is zoomed,
# it is similar to the INTER_NEAREST method.
INTER_AREA    = cv2.INTER_AREA

# A bicubic interpolation over 4x4 pixel neighborhood.
INTER_CUBIC   = cv2.INTER_CUBIC

# A Lanczos interpolation over 8x8 pixel neighborhood.
INTER_LANCZOS4 = cv2.INTER_LANCZOS4


def resize_stretch(im, out_size, interpolation=INTER_NEAREST):
    """
    * im: numpy array
    * out_size: tuple of integers (width, height)

    Return the stretched image im that fills out_size.
    """
    return cv2.resize(im, out_size, interpolation=interpolation)


def resize_contain(im, out_size, bgcolor=0, interpolation=INTER_NEAREST):
    """
    * im: numpy array
    * out_size: tuple of integers (width, height)
    * bgcolor: integer >= 0, used to fill the background
    * interpolation: interpolation method to use

    Return a scaled image im to the largest size that fits out_size.
    """
    im_height, im_width = im.shape[:2]
    im_ratio = im_width / float(im_height)
    out_ratio = out_size[0] / float(out_size[1])
    if im_ratio <= out_ratio:
        out_width = int(out_size[1] * im_ratio)
        out_height = out_size[1]
    else:
        out_width = out_size[0]
        out_height = int(out_size[0] / im_ratio)

    out = cv2.resize(im, (out_width, out_height), interpolation=interpolation)
    # If the output is too small, center it based on the specified size.
    if out.shape[1] < out_size[0]:
        extra_width = out_size[0] - out.shape[1]
        one_more = extra_width % 2
        out1 = numpy.zeros((out.shape[0], out_size[0]) + (out.shape[2:]), out.dtype)
        if bgcolor:
            out1.fill(bgcolor)
        out1[:, extra_width/2:-(extra_width/2 + one_more)] = out
        out = out1
    if out.shape[0] < out_size[1]:
        extra_height = out_size[1] - out.shape[0]
        one_more = extra_height % 2
        out1 = numpy.zeros((out_size[1], out.shape[1]) + (out.shape[2:]), out.dtype)
        if bgcolor:
            out1.fill(bgcolor)
        out1[extra_height/2:-(extra_height/2 + one_more), :] = out
        out = out1

    return out


def resize_cover(im, out_size, interpolation=INTER_NEAREST):
    """
    * im: numpy array
    * out_size: tuple of integers (width, height)
    * interpolation: interpolation method to use

    Return an image such that im covers out_size; edges are cropped if necessary.
    """
    im_height, im_width = im.shape[:2]
    im_ratio = im_width / float(im_height)
    out_ratio = out_size[0] / float(out_size[1])
    if im_ratio <= out_ratio:
        out_width = out_size[0]
        out_height = int(out_size[0] / im_ratio)
    else:
        out_width = int(out_size[1] * im_ratio)
        out_height = out_size[1]

    out = cv2.resize(im, (out_width, out_height), interpolation=interpolation)
    # If the output is too big, crop edges to match the specified size.
    if out.shape[1] > out_size[0]:
        extra_width = out.shape[1] - out_size[0]
        one_more = extra_width % 2
        out = out[:, extra_width/2:-(extra_width/2 + one_more)]
    if out.shape[0] > out_size[1]:
        extra_height = out.shape[0] - out_size[1]
        one_more = extra_height % 2
        out = out[extra_height/2:-(extra_height/2 + one_more), :]

    return out


def load_image(filename, area_size, binarize=0, padding=0):
    img = cv2.imread(filename)
    if img is None:
        return

    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    if padding:
        gray = util.pad(gray, padding)

    # Resize if needed.
    resize = max(gray.shape) > min(area_size)
    if resize:
        gray = util.resize(gray, area_size)

    # Binarize.
    if binarize is not None:
        if binarize >= 0:
            gray[gray <= binarize] = 0
            gray[gray > binarize] = 255
        else:
            # Use otsu.
            gray = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)[1]

    # Copy image to the center of area.
    area = numpy.zeros(area_size, numpy.uint8)
    util.copy_to_center(area, gray)

    return area



import time
import json
import struct
import logging
import logging.config
import collections
from functools import partial

import numpy
import zmq
from zmq.eventloop import zmqstream, ioloop

from .red import ReconnectingRedis
from .common import REDIS_KEYS, INPUT_STREAM, HDMI_INPUT_STREAM, OUTPUT_STREAM

__all__ = ['Application']

ioloop.install()


class Application(object):

    def __init__(self, name, config,
            setup_input=True, setup_output=True, setup_hdmi_input=True,
            verbose=False):
        """
        An user app instance.

        * name should be a string. It will be used to store app data on redis
          and to identify the app through the http server.
          Accessible through instance.name

        * config can be either a string or a dict, if it's a string
          json.load(open(config)) will be used.
          + "logging" will be passed to logging.config.dictConfig(..)
          + If "output_stream", "input_stream" or "hdmi_stream" keys are present,
          they must specify at least a "width" and "height" subkeys.
          + If a "redis" key is present it will be used to configure the redis
          client.
          + The "settings" key must be used for holding settings specific
          to the app that can be updated while it's running.
          The app is free to use any other keys.
          Accessible through instance.config

        * verbose should be a boolean.
          Accessible through instance.verbose

        * if setup_input is True a SUB socket for the input stream will be
          configured for the topic framework.INPUT_STREAM using
          config["input_stream"]. The same applies for setup_hdmi_input
          with topic frame.HDMI_INPUT_STREAM and config["hdmi_stream"].

        * if setup_output is True a PUB socket for the output stream will be
          configured using config["output_stream"]
        """
        config = config or {}
        self.name = name
        self.verbose = verbose
        if not isinstance(config, dict):
            # Load config from json file.
            config = json.load(open(config))
        self.config = config

        if 'logging' in config:
            logging.config.dictConfig(config['logging'])
        else:
            logging.basicConfig(
                    level=logging.DEBUG if verbose else logging.INFO,
                    format='%(asctime)s %(levelname)s %(name)s %(process)d %(message)s')

        self._log = logging.getLogger(__name__)

        # List of redis channels to subscribe.
        sub_channels = []

        self._input_callback = {}

        if 'input_stream' in config:
            input_width = config['input_stream']['width']
            input_height = config['input_stream']['height']
            self._bgr_shape = (input_height, input_width, 3)
            self._depth_shape = (input_height, input_width)
            sub_channels.append(REDIS_KEYS.SYS_INPUT_CHANNEL)
        else:
            setup_input = False
            self._bgr_shape = None
            self._depth_shape = None

        if 'hdmi_stream' in config:
            hdmi_width = config['hdmi_stream']['width']
            hdmi_height = config['hdmi_stream']['height']
            self._hdmi_shape = (hdmi_width, hdmi_height, 3)
            sub_channels.append(REDIS_KEYS.SYS_HDMI_CHANNEL)
        else:
            setup_hdmi_input = False
            self._hdmi_shape = None

        self._bgr_dtype = 'uint8'
        self._depth_dtype = 'uint16'
        self._hdmi_dtype = 'uint8'

        if 'output_stream' in config:
            output_width = config['output_stream']['width']
            output_height = config['output_stream']['height']
            self._bin_shape = (output_height, output_width)
            self._out_transition = config['output_stream'].get('transition', False)
            self._out_preview = config['output_stream'].get('preview', False)
            subtopic = config['output_stream'].get('subtopic')
            sub_channels.append(REDIS_KEYS.SYS_OUTPUT_CHANNEL)
        else:
            setup_output = False
            subtopic = None
            self._bin_shape = None
            self._out_transition = False
            self._out_preview = False
        self._bin_dtype = 'uint8'
        if self._out_transition and not self.name:
            raise ValueError("App name is required to use transitions")

        if self.name:
            subtopic = ':%s' % subtopic if subtopic else ''
            self._app_rkey = self.name + subtopic
            sub_channels.append(REDIS_KEYS.APP_CHANNEL + self._app_rkey)
        else:
            self._app_rkey = None

        self._ctx = zmq.Context()
        self._in_stream = {}
        self._sent = 0
        self._out_socket = None
        self._out_topic = None
        if setup_input:
            self.setup_input()
        if setup_hdmi_input:
            self.setup_input('hdmi_stream', HDMI_INPUT_STREAM)
        if setup_output:
            self.setup_output()

        self._cb_app_heartbeat = None
        self._periodic_callbacks = {}

        self._redis_sub_callback = None
        use_redis = True
        if 'redis' in config:
            use_redis = config['redis'].get('enabled', True)
        if not use_redis:
            self._red = None
            self._red_sub = None
        else:
            self._red = ReconnectingRedis()
            self._red_sub = ReconnectingRedis('pubsub')
            self._setup_redis(sub_channels)

    def setup_input(self, cfg='input_stream', topic=INPUT_STREAM, bind=False, watermark=0):
        """
        Configure a ZMQ PUB socket for the input stream.
        """
        if cfg in self._in_stream:
            # Already configured the input stream, nothing to do.
            return

        sock_address = self.config[cfg]['socket']
        in_socket = self._ctx.socket(zmq.SUB)
        in_socket.setsockopt(zmq.SUBSCRIBE, topic)
        if watermark > 0:
            in_socket.set_hwm(watermark)

        if bind:
            in_socket.bind(sock_address)
        else:
            in_socket.connect(sock_address)
        self._in_stream[cfg] = zmqstream.ZMQStream(in_socket)

        callback = getattr(self, '_input_callback_%s' % cfg, None)
        if callback is None:
            raise Exception('Implementation not available for %s configuration' % cfg)
        self._in_stream[cfg].on_recv(callback, copy=False)

        self._log.debug("SUB socket %s to %s, topic %s",
                'bound' if bind else 'connected', sock_address, topic)

    def set_input_callback(self, function, stream='input_stream'):
        """
        Define a callback to be invoked on messages received through
        the socket configured with setup_input.
        """
        self._input_callback[stream] = function

    def _input_callback_output_stream(self, msg):
        cb = self._input_callback.get('output_stream')
        if cb is None:
            return

        if len(msg) != 3:
            self._log.error('expected output message of size %d, got %d', 3, len(msg))
            return

        topic = msg[0].bytes
        subtopic = topic.split(':', 1)[1].rstrip(b'\x00') if ':' in topic else None
        frame_num = struct.unpack('i', msg[1])[0]
        binimage_frame = msg[2]

        bin_image = numpy.frombuffer(binimage_frame.bytes, dtype=self._bin_dtype)
        try:
            bin_image = bin_image.reshape(self._bin_shape)
        except ValueError:
            self._log.exception('ignoring bad frame')
            return

        cb(app=self, subtopic=subtopic, frame_num=frame_num, bin_image=bin_image)

    # Preview stream is processed the same way as the output stream.
    _input_callback_preview_stream = _input_callback_output_stream

    def _input_callback_input_transition_stream(self, msg):
        cb = self._input_callback.get('transition_stream')
        if cb is None:
            return

        if len(msg) != 4:
            self._log.error('expected transition message of size %d, got %d', 4, len(msg))
            return

        # topic = msg[0]  # unused.
        app_name = str(msg[1])[:-1]
        frame_num = struct.unpack('i', msg[2])[0]
        binimage_frame = msg[3]

        bin_image = numpy.frombuffer(binimage_frame.bytes, dtype=self._bin_dtype)
        try:
            bin_image = bin_image.reshape(self._bin_shape)
        except ValueError:
            self._log.exception('ignoring bad frame')
            return

        cb(app=self, frame_from=app_name, frame_num=frame_num, bin_image=bin_image)

    def _input_callback_input_stream(self, msg):
        cb = self._input_callback.get('input_stream')
        if cb is None:
            return

        if len(msg) != 4:
            self._log.error('expected input message of size %d, got %d', 4, len(msg))
            return

        # topic = msg[0]  # unused.
        frame_num = struct.unpack('i', msg[1])[0]
        depth_frame = msg[2]
        bgr_frame = msg[3]

        depth = numpy.frombuffer(depth_frame.bytes, dtype=self._depth_dtype)
        bgr = numpy.frombuffer(bgr_frame.bytes, dtype=self._bgr_dtype)
        try:
            depth = depth.reshape(self._depth_shape)
            bgr = bgr.reshape(self._bgr_shape)
        except ValueError:
            self._log.exception('ignoring bad frame')
            return

        cb(app=self, frame_num=frame_num, depth=depth, bgr=bgr)

    def _input_callback_hdmi_stream(self, msg):
        cb = self._input_callback.get('hdmi_stream')
        if cb is None:
            return

        if len(msg) != 3:
            self._log.error('expected hdmi message of size %d, got %d', 3, len(msg))
            return

        frame_num = struct.unpack('i', msg[1])[0]
        bgr_frame = msg[2]

        bgr = numpy.frombuffer(bgr_frame.bytes, dtype=self._hdmi_dtype)
        try:
            bgr = bgr.reshape(self._hdmi_shape)
        except ValueError:
            self._log.exception('ignoring bad frame')
            return

        cb(app=self, frame_num=frame_num, bgr=bgr)

    def setup_output(self, cfg='output_stream', bind=None):
        """
        Configure a ZMQ PUB socket for the output stream.
        """
        if self._out_socket:
            return

        sock_address = self.config[cfg]['socket']
        subtopic = self.config[cfg].get('subtopic')
        subtopic = ':%s' % subtopic if subtopic else ''

        self._out_socket = self._ctx.socket(zmq.PUB)
        self._out_topic = bytes('%s%s' % (OUTPUT_STREAM, subtopic))
        if bind is None:
            bind = (not self._out_transition) and (not self._out_preview)
        if bind:
            self._out_socket.bind(sock_address)
        else:
            self._out_socket.connect(sock_address)

        self._log.debug("PUB socket %s to %s, topic %s",
                'bound' if bind else 'connected', sock_address, self._out_topic)

    def send_output(self, result, topic=None):
        """
        Publish result using the output socket configured by setup_output.

        An integer is returned. This number indicates the number of frames
        sent so far, i.e. the number of send_output invocations, starting
        from 1.
        """
        if not self._out_socket:
            raise Exception("output stream not configured")

        self._sent += 1
        topic = self._out_topic if topic is None else bytes(topic)
        if self._out_transition:
            # When (potentially) using transitions, also send the app name.
            data = [b'%s\x00' % topic, b'%s\x00' % self.name,
                    struct.pack('i', self._sent), result.tobytes()]
        else:
            data = [b'%s\x00' % topic,
                    struct.pack('i', self._sent), result.tobytes()]
        self._out_socket.send_multipart(data, copy=False)
        return self._sent

    def notify(self, channel, data):
        """Send json.dumps(data) to a channel using redis pubsub."""
        if not self._red:
            raise Exception("redis not enabled")
        self._log.debug('Redis outgoing, channel: %s, data: %s', channel, data)
        return self._red.publish(channel, json.dumps(data))

    def set_redis_callback(self, function):
        """
        Define a callback to be invoked on messages received through
        the redis pubsub.
        """
        self._redis_sub_callback = function

    def add_periodic_callback(self, function, float_sec, start=False):
        """
        Schedule function to be called once each n seconds. The callback
        will receive this instance as its sole argument.

        The returned key can be used with stop_periodic_callback to stop
        the scheduling.
        """
        cb = ioloop.PeriodicCallback(
                lambda: function(self),
                float_sec * 1000)  # convert to ms
        key = id(cb)
        self._periodic_callbacks[key] = cb
        if start:
            cb.start()
        return key

    def stop_periodic_callback(self, key):
        """
        Stop a periodic callback created with add_periodic_callback.
        """
        cb = self._periodic_callbacks.pop(key)
        cb.stop()

    def call_later(self, float_sec, function, *args, **kwargs):
        """
        Schedule function to be called after float_sec seconds have passed.

        The returned handle can be used with cancel_call_later to cancel the
        future call.
        """
        handle = ioloop.IOLoop.current().call_later(float_sec, function, *args, **kwargs)
        return handle

    def cancel_call_later(self, handle):
        """
        Stop a scheduled call_later function from running.
        """
        ioloop.IOLoop.current().remove_timeout(handle)

    def stop_ioloop(self):
        """
        Stop the event loop.
        Note: call the cleanup function after this one if the app is quitting.
        """
        ioloop.IOLoop.current().stop()

    def run(self):
        """
        Start the registered callbacks and the event loop.
        If a name was specified, the app will be registered on
        redis if it's enabled.
        """
        if self.name and self._red is not None:
            self._cb_app_heartbeat = ioloop.PeriodicCallback(
                    self._app_heartbeat, 1000)
            self._cb_app_heartbeat.start()
        for cb in self._periodic_callbacks.values():
            if not cb.is_running():
                cb.start()
        ioloop.IOLoop.current().start()

    def cleanup(self):
        """
        Call this before quitting to stop registered callbacks and
        unregister the app from redis.
        """
        for cb in self._periodic_callbacks.values():
            cb.stop()
        if self._cb_app_heartbeat:
            self._cb_app_heartbeat.stop()
        if self.name and self._red is not None:
            # Unregister app from redis.
            self._red.hdel(REDIS_KEYS.APPS, self._app_rkey)

    def _app_heartbeat(self):
        # Update register on redis to indicate that the app is running well.
        self.config['timestamp'] = time.time()
        if self._red.is_connected():
            self._red.hset(REDIS_KEYS.APPS, self._app_rkey, json.dumps(self.config))

    def _setup_redis(self, channels):
        cfg = {'host': 'localhost', 'port': 6379, 'callback': self._on_redis_connect}
        if 'redis' in self.config:
            if 'host' in self.config['redis']:
                cfg['host'] = self.config['redis']['host']
            if 'port' in self.config['redis']:
                cfg['port'] = self.config['redis']['port']

        # Keep one redis connection for regular commands.
        self._red.connect(**cfg)
        # And another redis connection for pubsub.
        cb = partial(cfg.pop('callback'), sub_channels=channels)
        self._red_sub.connect(callback=cb, **cfg)

    def _on_redis_message(self, msg):
        if msg is None:
            return
        msg_type, msg_channel, content = msg
        if msg_type != 'message':
            return

        self._log.debug('Redis incoming, channel: %s, msg: %s', msg_channel, content)

        data = json.loads(content)
        if msg_channel.startswith(REDIS_KEYS.APP_CHANNEL):
            # Update app settings.
            _update_settings(self.config['settings'], data)
            self._app_heartbeat()
        elif msg_channel == REDIS_KEYS.SYS_INPUT_CHANNEL:
            # Input stream update.
            _update_settings(self.config['input_stream'], data)
            input_height = self.config['input_stream']['height']
            input_width = self.config['input_stream']['width']
            self._bgr_shape = (input_height, input_width, 3)
            self._depth_shape = (input_height, input_width)
        elif msg_channel == REDIS_KEYS.SYS_OUTPUT_CHANNEL:
            # Output stream update.
            _update_settings(self.config['output_stream'], data)
            self._bin_shape = (
                    self.config['output_stream']['height'],
                    self.config['output_stream']['width'])

        if self._redis_sub_callback:
            self._log.debug('Redis sub callback')
            self._redis_sub_callback(app=self, channel=msg_channel, update=data)

    def _on_redis_connect(self, sub_channels=None):
        if 'redis' in self.config:
            db = self.config['redis'].get('db')
            if db:
                self._red.select(db)
            pwd = self.config['redis'].get('auth')
            if pwd:
                self._red.auth(pwd)

        if sub_channels:
            self._red_sub.subscribe(sub_channels, self._on_redis_message)


def _update_settings(settings, new):
    # Restrict key updates to those that already exist and
    # enforce the values to be of same type.
    def update(orig_data, new_data):
        for key, new_value in new_data.items():
            if key in orig_data:
                if isinstance(new_value, collections.Mapping):
                    # Recurse on nested settings.
                    res = update(orig_data[key], new_value)
                    orig_data[key] = res
                else:
                    orig_data[key] = type(orig_data[key])(new_value)
        return orig_data
    update(settings, new)



import os
from setuptools import setup, Extension

import numpy

if os.getenv('USE_CYTHON', '0').lower() in ('0', 'f', 'false'):
    CYTHON_BUILD = None
    binarize_src = 'flipdisc/binarize.c'
else:
    from Cython.Distutils import build_ext as cython_build_ext
    CYTHON_BUILD = cython_build_ext
    binarize_src = 'flipdisc/binarize.pyx'
binarize_ext = Extension('flipdisc.binarize', [binarize_src])


def run_setup():
    extra = {
        'ext_modules': [binarize_ext],
        'include_dirs': [numpy.get_include()]
    }
    if CYTHON_BUILD:
        extra['cmdclass'] = {'build_ext': CYTHON_BUILD}

    setup(
        name='flipdisc',
        version='0.5.0',
        url='https://github.com/breakfastny/flipdisc-python-framework',
        packages=['flipdisc', 'flipdisc.framework'],
        package_data={
            'flipdisc.framework': ['*.json'],
        },
        setup_requires=['cffi>=1.0.0'],
        cffi_modules=['flipdisc/build_particle.py:ffibuilder'],
        install_requires=[
            'cffi>=1.0.0',
            'numpy==1.12.1',
            'pyzmq==16.0.2',
            'tornado==4.4.2',
            'toredis==0.1.2',
            'attrdict==2.0.0',
            'jsonschema==2.6.0',
        ],
        zip_safe=False,
        **extra
    )


run_setup()



from attrdict import AttrDict

REDIS_KEYS = AttrDict({
    # hashtable used to store all active apps.
    # key is an app name and the value is a json.dumps of all its settings.
    'APPS': 'fd:apps',

    # channel name used to send messages to specific apps.
    # When subscribing, use it as APP_CHANNEL + appname.
    'APP_CHANNEL': 'fd:app:',

    # channel name used to send messages to all apps that use input settings.
    'SYS_INPUT_CHANNEL': 'fd:system:input',
    # channel name used to send messages to all apps that use hdmi settings.
    'SYS_HDMI_CHANNEL': 'fd:system:hdmi',
    # channel name used to send messages to all apps that use output settings.
    'SYS_OUTPUT_CHANNEL': 'fd:system:output',
})

INPUT_STREAM = "IN_STREAM"
HDMI_INPUT_STREAM = "HDMI_STREAM"

OUTPUT_STREAM = "OUT_STREAM"



import os

from cffi import FFI

FDL_PATHS = []
if 'FDL_INCLUDE_DIR' in os.environ:
    FDL_PATHS.append(os.environ['FDL_INCLUDE_DIR'])
else:
    FDL_PATHS.append('/usr/local/include/fdl')
    FDL_PATHS.append('/usr/include/fdl')


def search_particle_h():
    for p in FDL_PATHS:
        check = os.path.join(p, 'particle.h')
        if os.path.exists(p) and os.path.exists(check):
            return check

    raise SystemExit('** ERROR: particle.h not found at %s' % (','.join(FDL_PATHS)))


ffibuilder = FFI()
ffibuilder.set_source('flipdisc._particle',
        '#include "fdl/particle.h"',
        libraries=['fparticle'])

raw_cdef = open(search_particle_h()).read()
# Skip includes and other things. This assumes the first definition
# in particle.h starts with a struct.
raw_start = raw_cdef.find('struct ')
raw_end = raw_cdef.find('#ifdef __cplusplus', raw_start + 1)
ffibuilder.cdef(raw_cdef[raw_start:raw_end])

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)
